"""Building engine to understand and process PDF files."""
import asyncio
import re
from collections import defaultdict

from fastapi import HTTPException
from google.cloud import documentai
from google.cloud import documentai_v1beta3 as docai_beta
from google.cloud.documentai_v1 import Document as docaiv1_document

from src.io import delete_folder_from_bucket, logger, upload_pdf_to_bucket
from src.postprocessing.postprocess_booking_confirmation import (
    postprocess_booking_confirmation,
)
from src.tms_formatting import format_TMS
from src.utils import find_new_data_fields, run_background_tasks


async def _process_pdf_w_docai(image_content, client, processor_name):
    """Process the PDF using Document AI.

    Args:
        image_content (bytes): The content of the PDF file as bytes.
        client: The Document AI client.
        processor_name (str): The name of the processor to be used.
                            e.g.: projects/{project_id}/locations/{location}/processor/{processor_id}

    Returns:
        The processed document.
    """
    # Load binary data
    raw_document = documentai.RawDocument(
        content=image_content, mime_type="application/pdf"
    )

    # Configure the process request
    request = documentai.ProcessRequest(
        name=processor_name,
        raw_document=raw_document,  # field_mask=field_mask
    )
    result = await client.process_document(request=request)

    return result.document


async def process_file_w_docai(params, image_content, client, processor_name):
    """
    Process a file using Document AI.

    Args:
        params (dict): The project parameters.
        image_content (bytes): The file to be processed. It can be bytes object.
        client: The Document AI client.
        processor_name (str): The name of the processor to be used.

    Returns:
        The processed document.

    Raises:
        ValueError: If the file is neither a path nor a bytes object.
    """
    result = None

    try:
        logger.info("Processing document...")
        result = await _process_pdf_w_docai(image_content, client, processor_name)
    except Exception as e:
        if e.reason == "PAGE_LIMIT_EXCEEDED":
            logger.warning(
                "Document contains more than 15 pages! Processing in batch method..."
            )
            # Process the document in batch method (offline processing)
            try:
                result = await _batch_process_pdf_w_docai(
                    params, image_content, client, processor_name
                )
            except Exception as batch_e:
                logger.error(f"Error processing document {batch_e}.")

        else:
            logger.error(f"Error processing document {e}.")

    return result


async def _batch_process_pdf_w_docai(
    params, image_content, client, processor_name, timeout=1200
):
    """Process the PDF using Document AI Batch Process API.

    Args:
        image_content (bytes): The content of the PDF file as bytes.
        client: The Document AI client.
        processor_name (str): The name of the processor to be used.
                            e.g.: projects/{project_id}/locations/{location}/processor/{processor_id}
        timeout (int, optional): The timeout in seconds. Defaults to 1200.

    Returns:
        The processed document.
    """
    # Upload the PDF to GCS bucket
    gcs_input_uri, storage_client = upload_pdf_to_bucket(
        params, image_content, "temp.pdf"
    )

    gcs_document = documentai.GcsDocument(
        gcs_uri=gcs_input_uri, mime_type="application/pdf"
    )
    # Load GCS Input URI into a List of document files
    input_config = documentai.BatchDocumentsInputConfig(
        gcs_documents=documentai.GcsDocuments(documents=[gcs_document])
    )

    # Cloud Storage URI for the Output Directory
    # This must end with a trailing forward slash `/`
    destination_uri = f"gs://{params['doc_ai_bucket_batch_output']}/"  # noqa
    gcs_output_config = documentai.DocumentOutputConfig.GcsOutputConfig(
        gcs_uri=destination_uri, field_mask="entities"
    )

    # Where to write results
    output_config = documentai.DocumentOutputConfig(gcs_output_config=gcs_output_config)

    # The full resource name of the processor
    request = documentai.BatchProcessRequest(
        name=processor_name,
        input_documents=input_config,
        document_output_config=output_config,
    )

    # BatchProcess returns a Long Running Operation (LRO)
    logger.info("Processing document in batch mode...")
    operation = await client.batch_process_documents(request)

    try:
        # Wait for the operation to finish
        logger.info(f"Waiting for operation {operation.operation.name} to complete...")
        await operation.result(timeout=timeout)
    # Catch exception when operation doesn't finish before timeout
    except Exception as e:
        logger.error(e)

    # Once the operation is complete,
    # get output document information from operation metadata
    metadata = documentai.BatchProcessMetadata(operation.metadata)

    if metadata.state != documentai.BatchProcessMetadata.State.SUCCEEDED:
        raise ValueError(f"Batch Process Failed: {metadata.state}")

    # One process per Input Document
    for process in metadata.individual_process_statuses:
        # The Cloud Storage API requires the bucket name and URI prefix separately
        matches = re.match(r"gs://(.*?)/(.*)", process.output_gcs_destination)
        if not matches:
            logger.warning(
                "Could not parse output GCS destination:",
                process.output_gcs_destination,
            )
            continue

        output_bucket, output_prefix = matches.groups()

        # Get List of Document Objects from the Output Bucket
        output_blobs = storage_client.list_blobs(output_bucket, prefix=output_prefix)

        # Document AI may output multiple JSON files per source file
        # Selecting only the first output file as of now. No particular reason
        for blob in output_blobs:
            # Document AI should only output JSON files to GCS
            if ".json" not in blob.name:
                logger.warning(
                    f"Skipping non-supported file: {blob.name} - Mimetype: {blob.content_type}"
                )
                continue

            # Download JSON File as bytes object and convert to Document Object
            result_document = documentai.Document.from_json(
                blob.download_as_bytes(), ignore_unknown_fields=True
            )

            # Delete the temporary file and the output file from the bucket
            delete_folder_from_bucket(params["doc_ai_bucket_batch_input"], "temp.pdf")
            delete_folder_from_bucket(output_bucket, output_prefix)
            logger.info("Batch Process Completed!")

            return result_document


async def split_pdf(params, file_content, processor_client):
    """Split the PDF into sub-documents based on the detected document types."""
    # Get the splitter model name
    location = params.get("g_location")
    project = params.get("g_ai_project_id")
    processor = params.get("g_splitter_model_id")
    processor_name = processor_client.processor_path(
        project=project, location=location, processor=processor
    )
    result = await _process_pdf_w_docai(file_content, processor_client, processor_name)
    sub_documents = [
        {
            "page_range": [
                int(page_bounders.page)
                for page_bounders in sub_document.page_anchor.page_refs
            ],
            "detected_doc_type": sub_document.type_,
        }
        for sub_document in result.entities
    ]
    return sub_documents


async def extract_data_from_pdf(
    params, input_doc_type, file_content, processor_client, schema_client, embed_manager
):
    """Extract data from the PDF file."""
    processor_name = params["data_extractor_processor_names"].get(input_doc_type)

    if input_doc_type == "bookingConfirmation_test":
        processor_name = params["data_extractor_processor_names"].get("BC_mlg")

    confidence = {}
    if (
        input_doc_type in params["models"]
        and "confidence" in params["models"][input_doc_type]
    ):
        confidence = params["models"][input_doc_type]["confidence"]

    if not processor_name:
        supported_doc_types = list(params["data_extractor_processor_names"].keys())
        raise HTTPException(
            status_code=400,
            detail=f"Unsupported document type {input_doc_type}. Supported document types are: {supported_doc_types}",  # noqa: E501
        )

    # Check if the processor schema is updated
    # ie. if the document type's schema is updated compared to the schema of best trained model/ processor version
    new_data_fields, zero_shot_model_version = await find_new_data_fields(
        params, processor_name, processor_client, schema_client
    )
    if new_data_fields:
        # Predict using the Zero-shot model for the new schema
        result, zero_shot_result = await asyncio.gather(
            process_file_w_docai(
                params, file_content, processor_client, processor_name
            ),
            process_file_w_docai(
                params, file_content, processor_client, zero_shot_model_version
            ),
        )

        # Append the new header fields to the old schema
        result.entities.extend(
            entity
            for entity in zero_shot_result.entities
            if entity.type_ in new_data_fields
        )
    else:
        result = await process_file_w_docai(
            params, file_content, processor_client, processor_name
        )
        # TODO: Deleted the deprecated fields from the schema
        # Update the "find_new_data_fields" function to return the deprecated fields too

    # Create an entity object to store the result in gcs
    result_for_store = docaiv1_document.to_json(result)

    aggregated_data = defaultdict(list)

    for entity in result.entities:
        child_entities = entity.properties
        if child_entities:
            for child in child_entities:
                if child.confidence >= confidence.get(child.type_, 0):
                    value = format_TMS(child, embed_manager)
                    aggregated_data[child.type_].append(value)
        elif entity.confidence >= confidence.get(entity.type_, 0):
            value = format_TMS(entity, embed_manager)
            aggregated_data[entity.type_].append(value)

    aggregated_data = await validate_based_on_schema(
        aggregated_data, processor_name, schema_client
    )

    logger.info("Data Extraction completed successfully")

    # Call postprocessing for Multi Leg
    if (
        input_doc_type == "bookingConfirmation"
        or input_doc_type == "bookingConfirmation_test"
    ):
        aggregated_data = postprocess_booking_confirmation(aggregated_data)
        logger.info("Transport Legs assembled successfully")

    response = await processor_client.get_processor(name=processor_name)
    processor_version = response.default_processor_version.split("/")[-1]

    logger.info(
        f"Processor & it's version used for current request: {response.display_name} - {processor_version}"
    )

    return aggregated_data, result_for_store, processor_version


async def validate_based_on_schema(extracted_raw_data, processor_name, schema_client):
    """Validate the extracted data based on the schema."""
    schema_request = docai_beta.GetDatasetSchemaRequest(
        name=f"{processor_name}/dataset/datasetSchema"
    )
    schema_response = await schema_client.get_dataset_schema(request=schema_request)
    # schema_response.document_schema.entity_types contains 2 elements:
    # One for entities at document level
    # One for entities at line item level
    schemas = schema_response.document_schema.entity_types
    schema_header_fields = schemas[0].properties

    result = dict(extracted_raw_data)
    for data_field in schema_header_fields:
        if "once" in data_field.occurrence_type.name.lower():
            if data_field.name in result:
                result[data_field.name] = result[data_field.name][0]
    return result


async def data_extraction_manual_flow(
    params,
    file_content,
    mime_type,
    meta,
    processor_client,
    schema_client,
    embed_manager,
):
    """
    Process a PDF file and extract data from it.

    Args:
        params (dict): Parameters for the data extraction process.
        file_content (bytes): The content of the PDF file to process.
        mime_type (str): The MIME type of the document.
        meta (DocumentMeta): Metadata associated with the document.
        processor_client (DocumentProcessorClient): Client for the Document AI processor.
        schema_client (DocumentSchemaClient): Client for the Document AI schema.
        embed_manager (EmbeddingsManager): Manager for embeddings.

    Returns:
        dict: A dictionary containing the processed document information.

    Raises:
        Refer to reasons in 400 error response examples.
    """
    # Validate the file type
    if mime_type != "application/pdf":
        raise HTTPException(
            status_code=400, detail="Invalid file type. Please upload a PDF file."
        )

    # Extract data from the PDF file
    extracted_data, store_data, processor_version = await extract_data_from_pdf(
        params,
        input_doc_type=meta.documentTypeCode,
        file_content=file_content,
        processor_client=processor_client,
        schema_client=schema_client,
        embed_manager=embed_manager,
    )
    # Create the result dictionary with the extracted data
    result = {
        "id": meta.id,
        "documentTypeCode": meta.documentTypeCode,
        "data": extracted_data,
        "processor_version": processor_version,
    }

    # Schedule background tasks without using FastAPI's BackgroundTasks
    asyncio.create_task(
        run_background_tasks(
            params,
            meta.id,
            meta.documentTypeCode,
            extracted_data,
            store_data,
            processor_version,
            mime_type,
        )
    )

    return result
